@@ -8,20 +8,10 @@ from .strategies import (
     HalvingSearch,
     HyperbandSearch,
     PBTSearch,
+    Strategies,
 )
 from .decorator import hyperopt
 
-Strategies = {
-    "Random": RandomSearch,
-    "Grid": GridSearch,
-    "SMBO": SMBOSearch,
-    "Nevergrad": NevergradSearch,
-    "Coordinate": CoordinateSearch,
-    "Halving": HalvingSearch,
-    "Hyperband": HyperbandSearch,
-    "PBT": PBTSearch,
-}
-
 __all__ = [
     "__version__",
     "RandomSearch",
