@@ -1,4 +1,4 @@
-from typing import Union
+from typing import Optional
 from ..space import HyperSpace
 import nevergrad as ng
 
@@ -6,14 +6,30 @@ import nevergrad as ng
 class NevergradSpace(HyperSpace):
     def __init__(
         self,
-        real: Union[dict, None] = None,
-        integer: Union[dict, None] = None,
-        categorical: Union[dict, None] = None,
+        real: Optional[dict] = None,
+        integer: Optional[dict] = None,
+        categorical: Optional[dict] = None,
     ):
-        """For SMBO-based hyperopt generate spaces with skopt classes"""
+        """Search hyperparameter space for FAIR's nevergrad.
+        Reference: https://facebookresearch.github.io/nevergrad/
+
+        Args:
+            real (Optional[dict], optional):
+                Dictionary of real-valued search variables & their prior.
+                E.g. {"lrate": {"begin": 0.1, "end": 0.5, "prior": "uniform"}}
+                Defaults to None.
+            integer (Optional[dict], optional):
+                Dictionary of integer-valued search variables & their prior.
+                E.g. {"batch_size": {"begin": 1, "end": 5, "prior": "log-uniform"}}
+                Defaults to None.
+            categorical (Optional[dict], optional):
+                Dictionary of categorical-valued search variables.
+                E.g. {"arch": ["mlp", "cnn"]}
+                Defaults to None.
+        """
         HyperSpace.__init__(self, real, integer, categorical)
 
-    def check(self):
+    def check(self) -> None:
         """Check that all inputs are provided correctly."""
         if self.real is not None:
             real_keys = ["begin", "end", "prior"]
@@ -40,8 +56,8 @@ class NevergradSpace(HyperSpace):
                 if type(v) is not list:
                     self.categorical[k] = [v]
 
-    def construct(self):
-        """Setup/construct the search space."""
+    def construct(self) -> None:
+        """Setup/construct the search space for nevergrad."""
         param_dict = {}
         if self.categorical is not None:
             for k, v in self.categorical.items():
