@@ -1,4 +1,4 @@
-from typing import Union
+from typing import Optional
 import numpy as np
 from ..space import HyperSpace
 
@@ -6,14 +6,29 @@ from ..space import HyperSpace
 class RandomSpace(HyperSpace):
     def __init__(
         self,
-        real: Union[dict, None] = None,
-        integer: Union[dict, None] = None,
-        categorical: Union[dict, None] = None,
+        real: Optional[dict] = None,
+        integer: Optional[dict] = None,
+        categorical: Optional[dict] = None,
     ):
-        """For random hyperopt generate list/1d-line range to sample from."""
+        """Random search hyperparameter space with desired priors.
+
+        Args:
+            real (Optional[dict], optional):
+                Dictionary of real-valued search variables & their priors.
+                E.g. {"lrate": {"begin": 0.1, "end": 0.5, "prior": "log-uniform"}}
+                Defaults to None.
+            integer (Optional[dict], optional):
+                Dictionary of integer-valued search variables & their priors.
+                E.g. {"batch_size": {"begin": 1, "end": 5, "bins": "uniform"}}
+                Defaults to None.
+            categorical (Optional[dict], optional):
+                Dictionary of categorical-valued search variables.
+                E.g. {"arch": ["mlp", "cnn"]}
+                Defaults to None.
+        """
         HyperSpace.__init__(self, real, integer, categorical)
 
-    def check(self):
+    def check(self) -> None:
         """Check that all inputs are provided correctly."""
         if self.real is not None:
             real_keys = ["begin", "end", "prior"]
@@ -40,8 +55,8 @@ class RandomSpace(HyperSpace):
                 if type(v) is not list:
                     self.categorical[k] = [v]
 
-    def construct(self):
-        """Setup/construct the search space."""
+    def construct(self) -> None:
+        """Setup/construct the random search space."""
         param_range = {}
         if self.categorical is not None:
             for k, v in self.categorical.items():
@@ -59,13 +74,19 @@ class RandomSpace(HyperSpace):
             for k, v in self.integer.items():
                 param_range[k] = {
                     "value_type": "integer",
-                    "values": np.arange(int(v["begin"]), int(v["end"]) + 1, 1).tolist(),
+                    "values": np.arange(
+                        int(v["begin"]), int(v["end"]) + 1, 1
+                    ).tolist(),
                     "prior": v["prior"],
                 }
         self.param_range = param_range
 
-    def sample(self):
-        """'Sample' from the hyperparameter space."""
+    def sample(self) -> dict:
+        """Sample from the random hyperparameter space.
+
+        Returns:
+            dict: Randomly sampled parameter configuration dictionary.
+        """
         proposal_params = {}
         # Sample the parameters individually at random from the ranges
         for p_name, p_range in self.param_range.items():
@@ -76,7 +97,8 @@ class RandomSpace(HyperSpace):
                     eval_param = int(np.random.choice(p_range["values"]))
                 elif p_range["prior"] == "log-uniform":
                     x = np.random.uniform(
-                        np.log(p_range["values"][0]), np.log(p_range["values"][-1])
+                        np.log(p_range["values"][0]),
+                        np.log(p_range["values"][-1]),
                     )
                     eval_param = int(np.exp(x))
             elif p_range["value_type"] == "real":
@@ -84,7 +106,8 @@ class RandomSpace(HyperSpace):
                     eval_param = np.random.uniform(*p_range["values"])
                 elif p_range["prior"] == "log-uniform":
                     x = np.random.uniform(
-                        np.log(p_range["values"][0]), np.log(p_range["values"][1])
+                        np.log(p_range["values"][0]),
+                        np.log(p_range["values"][1]),
                     )
                     eval_param = np.exp(x)
             proposal_params[p_name] = eval_param
