@@ -1,5 +1,15 @@
-from .helpers import load_log, save_log, load_strategy, save_strategy, write_configs
-from .plotting import visualize_2D_grid
+from .helpers import (
+    save_yaml,
+    load_log,
+    save_log,
+    load_strategy,
+    save_strategy,
+    write_configs,
+    flatten_config,
+    unflatten_config,
+    merge_config_dicts,
+)
+from .plotting import visualize_2D_grid, load_search_log
 from .comms import (
     welcome_message,
     update_message,
@@ -15,12 +25,17 @@ from .comms import (
 
 
 __all__ = [
+    "save_yaml",
     "load_log",
     "save_log",
     "load_strategy",
     "save_strategy",
     "write_configs",
+    "flatten_config",
+    "unflatten_config",
+    "merge_config_dicts",
     "visualize_2D_grid",
+    "load_search_log",
     "welcome_message",
     "update_message",
     "ranking_message",
